/**
 * @file task_gps.c
 * @author CuBeatSystems
 * @author Shinichiro Nakamura
 */

/*
 *  TOPPERS Realtime System Sample Program - LPCXpresso GPS Clock
 *
 *  Copyright (C) 2010-2013 CuBeatSystems
 *  Copyright (C) 2010-2013 Shinichiro Nakamura
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include <itron.h>
#include <target_syssvc.h>
#include <syssvc/serial.h>
#include "kernel_cfg.h"
#include "task_gps.h"
#include "task_display.h"
#include "nmea.h"
#include "ntlibc.h"
#include "gt_720f.h"

#define GT720F_CUSTOM_SETUP     (0)
#define GPS_PORTID              (SIO4_PORTID)
#define CALLBACK_RETURN_VALUE   (-1)

/**
 * @brief データフォーマット構造体。
 * @details データ構造とコールバック関数の関係を表現している。
 */
typedef struct {
    char *header;
    int (*cbfunc)(int argc, char **argv, void *extobj);
} tagfunc_table_t;

static int cb_gga(int argc, char **argv, void *extobj)
{
    int hh = ((argv[1][0] - '0') * 10) + ((argv[1][1] - '0') * 1);
    int mm = ((argv[1][2] - '0') * 10) + ((argv[1][3] - '0') * 1);
    int ss = ((argv[1][4] - '0') * 10) + ((argv[1][5] - '0') * 1);
    hh += 9;
    if (24 <= hh) {
        hh -= 24;
    }
    tskapi_display_update(
            '0' + (hh / 10),
            '0' + (hh % 10),
            '0' + (mm / 10),
            '0' + (mm % 10),
            '0' + (ss / 10),
            '0' + (ss % 10),
            0);
    return CALLBACK_RETURN_VALUE;
}

static int cb_gll(int argc, char **argv, void *extobj)
{
    return CALLBACK_RETURN_VALUE;
}

static int cb_gsa(int argc, char **argv, void *extobj)
{
    return CALLBACK_RETURN_VALUE;
}

static int cb_gsv(int argc, char **argv, void *extobj)
{
    return CALLBACK_RETURN_VALUE;
}

static int cb_rmc(int argc, char **argv, void *extobj)
{
    return CALLBACK_RETURN_VALUE;
}

static int cb_vtg(int argc, char **argv, void *extobj)
{
    return CALLBACK_RETURN_VALUE;
}

static int cb_zda(int argc, char **argv, void *extobj)
{
    return CALLBACK_RETURN_VALUE;
}

static int nmea_getc(void)
{
    char_t c;
    serial_rea_dat(GPS_PORTID, &c, 1);
    return c;
}

static const tagfunc_table_t tagfunclist[] = {
    { "$GPGGA", cb_gga },
    { "$GPGLL", cb_gll },
    { "$GPGSA", cb_gsa },
    { "$GPGSV", cb_gsv },
    { "$GPRMC", cb_rmc },
    { "$GPVTG", cb_vtg },
    { "$GPZDA", cb_zda },
    {     NULL, NULL   },
};

static int nmea_cbfunc(int argc, char **argv, void *extobj)
{
    tagfunc_table_t *p = (tagfunc_table_t *)&tagfunclist[0];
    while (p->header != NULL) {
        if (ntlibc_strcmp(argv[0], p->header) == 0) {
            if (p->cbfunc != NULL) {
                return p->cbfunc(argc, argv, NULL);
            }
        }
        p++;
    }
    syslog(LOG_NOTICE, "Unknown header: %s(%d)", argv[0], argc);
    return CALLBACK_RETURN_VALUE;
}

/**
 * @brief GPS用リード関数。
 *
 * @param buf バッファ。
 * @param len 読み込み長さ。
 *
 * @retval 0 成功。
 * @retval !0 失敗。
 */
static int gps_read(unsigned char *buf, int len)
{
    return serial_rea_dat(GPS_PORTID, (char_t *)buf, len);
}

/**
 * @brief GPS用ライト関数。
 *
 * @param buf バッファ。
 * @param len 読み込み長さ。
 *
 * @retval 0 成功。
 * @retval !0 失敗。
 */
static int gps_write(unsigned char *buf, int len)
{
    return serial_wri_dat(GPS_PORTID, (const char_t *)buf, len);
}

/**
 * @brief クロックタスクの本体。
 * @param exinf タスクパラメータ。
 */
void task_gps(intptr_t exinf)
{
    /*
     * 開始処理。
     */
    syslog(LOG_NOTICE, "GPS task initialized.");
    serial_opn_por(GPS_PORTID);
    serial_ctl_por(GPS_PORTID, 0);

#if (GT720F_CUSTOM_SETUP != 0)
    unsigned char gps_gga = 1;
    unsigned char gps_gsa = 0;
    unsigned char gps_gsv = 0;
    unsigned char gps_gll = 0;
    unsigned char gps_rmc = 0;
    unsigned char gps_vtg = 0;
    unsigned char gps_zda = 0;

    gt_720f_t gt_720f;
    GT_720F_SETUP(&gt_720f, gps_write, gps_read);

    gt_720f_configure_message_type(&gt_720f, MessageTypeNoOutput);
    syslog(LOG_NOTICE, "GT-720F:Message type changed to 'No Output'.");
    tslp_tsk(1000);

    gt_720f_configure_nmea_message_interval(
            &gt_720f,
            gps_gga,
            gps_gsa,
            gps_gsv,
            gps_gll,
            gps_rmc,
            gps_vtg,
            gps_zda);
    syslog(LOG_NOTICE, "GT-720F:NMEA message type changed.");

    gt_720f_configure_message_type(&gt_720f, MessageTypeNMEA);
    syslog(LOG_NOTICE, "GT-720F:Message type changed to 'NMEA'.");
#endif

    /*
     * タスクループ。
     */
    while (1) {
        nmea_parse(nmea_getc, nmea_cbfunc, NULL);
    }

    /*
     * 終了処理。
     */
    serial_cls_por(GPS_PORTID);
    syslog(LOG_NOTICE, "GPS task terminated.");
    ext_ker();
}

